@@ -1,4 +1,4 @@
-# Copyright 2016-2021 The NATS Authors
+# Copyright 2016-2022 The NATS Authors
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
 # You may obtain a copy of the License at
@@ -56,7 +56,7 @@ from .subscription import (
     Subscription,
 )
 
-__version__ = '2.0.0'
+__version__ = '2.1.0'
 __lang__ = 'python3'
 _logger = logging.getLogger(__name__)
 PROTOCOL = 1
@@ -73,10 +73,9 @@ EMPTY = ""
 
 PING_PROTO = PING_OP + _CRLF_
 PONG_PROTO = PONG_OP + _CRLF_
-INBOX_PREFIX = bytearray(b'_INBOX.')
-INBOX_PREFIX_LEN = len(INBOX_PREFIX) + 22 + 1
+DEFAULT_INBOX_PREFIX = b'_INBOX'
 
-DEFAULT_PENDING_SIZE = 1024 * 1024
+DEFAULT_PENDING_SIZE = 2 * 1024 * 1024
 DEFAULT_BUFFER_SIZE = 32768
 DEFAULT_RECONNECT_TIME_WAIT = 2  # in seconds
 DEFAULT_MAX_RECONNECT_ATTEMPTS = 60
@@ -84,6 +83,7 @@ DEFAULT_PING_INTERVAL = 120  # in seconds
 DEFAULT_MAX_OUTSTANDING_PINGS = 2
 DEFAULT_MAX_PAYLOAD_SIZE = 1048576
 DEFAULT_MAX_FLUSHER_QUEUE_SIZE = 1024
+DEFAULT_FLUSH_TIMEOUT = 10  # in seconds
 DEFAULT_CONNECT_TIMEOUT = 2  # in seconds
 DEFAULT_DRAIN_TIMEOUT = 30  # in seconds
 MAX_CONTROL_LINE_SIZE = 1024
@@ -178,16 +178,26 @@ class Client:
         self._subs: Dict[int, Subscription] = {}
         self._status: int = Client.DISCONNECTED
         self._ps: Parser = Parser(self)
+
+        # pending queue of commands that will be flushed to the server.
         self._pending: List[bytes] = []
+
+        # current size of pending data in total.
         self._pending_data_size: int = 0
+
+        # max pending size is the maximum size of the data that can be buffered.
+        self._max_pending_size: int = 0
+
         self._flush_queue: Optional["asyncio.Queue[None]"] = None
         self._flusher_task: Optional[asyncio.Task] = None
+        self._flush_timeout: float = 0
         self._hdr_parser: BytesParser = BytesParser()
 
         # New style request/response
         self._resp_map: Dict[str, asyncio.Future] = {}
         self._resp_sub_prefix: Optional[bytearray] = None
         self._nuid = NUID()
+        self._inbox_prefix = bytearray(DEFAULT_INBOX_PREFIX)
 
         # NKEYS support
         #
@@ -219,7 +229,7 @@ class Client:
 
     async def connect(
         self,
-        servers: List[str] = ["nats://127.0.0.1:4222"],
+        servers: Union[str, List[str]] = ["nats://localhost:4222"],
         error_cb: Optional[ErrorCallback] = None,
         disconnected_cb: Optional[Callback] = None,
         closed_cb: Optional[Callback] = None,
@@ -247,6 +257,9 @@ class Client:
         user_jwt_cb: Optional[JWTCallback] = None,
         user_credentials: Optional[Credentials] = None,
         nkeys_seed: Optional[str] = None,
+        inbox_prefix: Union[str, bytes] = DEFAULT_INBOX_PREFIX,
+        pending_size: int = DEFAULT_PENDING_SIZE,
+        flush_timeout: Optional[float] = None,
     ) -> None:
         """
         Establishes a connection to NATS.
@@ -257,6 +270,8 @@ class Client:
         :param disconnected_cb: Callback to report disconnection from NATS.
         :param closed_cb: Callback to report when client stops reconnection to NATS.
         :param discovered_server_cb: Callback to report when a new server joins the cluster.
+        :param pending_size: Max size of the pending buffer for publishing commands.
+        :param flush_timeout: Max duration to wait for a forced flush to occur.
 
         Connecting setting all callbacks::
 
@@ -347,6 +362,12 @@ class Client:
         self._reconnected_cb = reconnected_cb
         self._disconnected_cb = disconnected_cb
 
+        # Custom inbox prefix
+        if isinstance(inbox_prefix, str):
+            inbox_prefix = inbox_prefix.encode()
+        assert isinstance(inbox_prefix, bytes)
+        self._inbox_prefix = bytearray(inbox_prefix)
+
         # NKEYS support
         self._signature_cb = signature_cb
         self._user_jwt_cb = user_jwt_cb
@@ -378,9 +399,15 @@ class Client:
         if self._user_credentials is not None or self._nkeys_seed is not None:
             self._setup_nkeys_connect()
 
-        # Queue used to trigger flushes to the socket
+        # Queue used to trigger flushes to the socket.
         self._flush_queue = asyncio.Queue(maxsize=flusher_queue_size)
 
+        # Max size of buffer used for flushing commands to the server.
+        self._max_pending_size = pending_size
+
+        # Max duration for a force flush (happens when a buffer is full).
+        self._flush_timeout = flush_timeout
+
         if self.options["dont_randomize"] is False:
             shuffle(self._server_pool)
 
@@ -549,8 +576,7 @@ class Client:
             return
         self._status = Client.CLOSED
 
-        # Kick the flusher once again so it breaks
-        # and avoid pending futures.
+        # Kick the flusher once again so that Task breaks and avoid pending futures.
         await self._flush_pending()
 
         if self._reading_task is not None and not self._reading_task.cancelled(
@@ -666,7 +692,7 @@ class Client:
         subject: str,
         payload: bytes = b'',
         reply: str = '',
-        headers: Optional[Dict[str, Any]] = None
+        headers: Optional[Dict[str, str]] = None
     ) -> None:
         """
         Publishes a NATS message.
@@ -719,6 +745,12 @@ class Client:
             raise errors.ConnectionDrainingError
 
         payload_size = len(payload)
+        if not self.is_connected:
+            if self._max_pending_size <= 0 or payload_size + self._pending_data_size > self._max_pending_size:
+                # Cannot publish during a reconnection when the buffering is disabled,
+                # or if pending buffer is already full.
+                raise errors.OutboundBufferLimitError
+
         if payload_size > self._max_payload:
             raise errors.MaxPayloadError
         await self._send_publish(
@@ -822,7 +854,8 @@ class Client:
     async def _init_request_sub(self) -> None:
         self._resp_map = {}
 
-        self._resp_sub_prefix = INBOX_PREFIX[:]
+        self._resp_sub_prefix = self._inbox_prefix[:]
+        self._resp_sub_prefix.extend(b'.')
         self._resp_sub_prefix.extend(self._nuid.next())
         self._resp_sub_prefix.extend(b'.')
         resp_mux_subject = self._resp_sub_prefix[:]
@@ -832,7 +865,7 @@ class Client:
         )
 
     async def _request_sub_callback(self, msg: Msg) -> None:
-        token = msg.subject[INBOX_PREFIX_LEN:]
+        token = msg.subject[len(self._inbox_prefix) + 22 + 2:]
         try:
             fut = self._resp_map.get(token)
             if not fut:
@@ -916,7 +949,8 @@ class Client:
            nc.publish('broadcast', b'', reply=inbox)
            msg = sub.next_msg()
         """
-        next_inbox = INBOX_PREFIX[:]
+        next_inbox = self._inbox_prefix[:]
+        next_inbox.extend(b'.')
         next_inbox.extend(self._nuid.next())
         return next_inbox.decode()
 
@@ -948,12 +982,12 @@ class Client:
             future.cancel()
             raise errors.TimeoutError
 
-    async def _send_unsubscribe(self, sid: int, limit: int = 1) -> None:
+    async def _send_unsubscribe(self, sid: int, limit: int = 0) -> None:
         unsub_cmd = prot_command.unsub_cmd(sid, limit)
         await self._send_command(unsub_cmd)
         await self._flush_pending()
 
-    async def flush(self, timeout: int = 10) -> None:
+    async def flush(self, timeout: int = DEFAULT_FLUSH_TIMEOUT) -> None:
         """
         Sends a ping to the server expecting a pong back ensuring
         what we have written so far has made it to the server and
@@ -973,12 +1007,12 @@ class Client:
             await asyncio.wait_for(future, timeout)
         except asyncio.TimeoutError:
             future.cancel()
-            raise errors.TimeoutError
+            raise errors.FlushTimeoutError
 
     @property
-    def connected_url(self) -> Optional[str]:
+    def connected_url(self) -> Optional[ParseResult]:
         if self._current_server and self.is_connected:
-            return str(self._current_server.uri)
+            return self._current_server.uri
         return None
 
     @property
@@ -1054,17 +1088,30 @@ class Client:
         else:
             self._pending.append(cmd)
         self._pending_data_size += len(cmd)
-        if self._pending_data_size > DEFAULT_PENDING_SIZE:
-            await self._flush_pending()
+        if self._max_pending_size > 0 and self._pending_data_size > self._max_pending_size:
+            # Only flush force timeout on publish
+            await self._flush_pending(force_flush=True)
 
-    async def _flush_pending(self) -> None:
+    async def _flush_pending(
+        self,
+        force_flush: bool = False,
+    ) -> None:
         assert self._flush_queue, "Client.connect must be called first"
         try:
-            # kick the flusher!
-            await self._flush_queue.put(None)
-
+            future = asyncio.Future()
             if not self.is_connected:
-                return
+                future.set_result(None)
+                return future
+
+            # kick the flusher!
+            await self._flush_queue.put(future)
+
+            if force_flush:
+                try:
+                    await asyncio.wait_for(future, self._flush_timeout)
+                except asyncio.TimeoutError:
+                    # Report to the async callback that there was a timeout.
+                    await self._error_cb(errors.FlushTimeoutError())
 
         except asyncio.CancelledError:
             pass
@@ -1074,11 +1121,11 @@ class Client:
             try:
                 if "nats://" in connect_url or "tls://" in connect_url:
                     # Closer to how the Go client handles this.
-                    # e.g. nats://127.0.0.1:4222
+                    # e.g. nats://localhost:4222
                     uri = urlparse(connect_url)
                 elif ":" in connect_url:
                     # Expand the scheme for the user
-                    # e.g. 127.0.0.1:4222
+                    # e.g. localhost:4222
                     uri = urlparse(f"nats://{connect_url}")
                 else:
                     # Just use the endpoint with the default NATS port.
@@ -1274,7 +1321,7 @@ class Client:
                     max_msgs = 0
                     if sub._max_msgs > 0:
                         # If we already hit the message limit, remove the subscription and don't
-                        # resubscribe
+                        # resubscribe.
                         if sub._received >= sub._max_msgs:
                             subs_to_remove.append(sid)
                             continue
@@ -1298,6 +1345,7 @@ class Client:
                 # Flush pending data before continuing in connected status.
                 # FIXME: Could use future here and wait for an error result
                 # to bail earlier in case there are errors in the connection.
+                # await self._flush_pending(force_flush=True)
                 await self._flush_pending()
                 self._status = Client.CONNECTED
                 await self.flush()
@@ -1425,9 +1473,11 @@ class Client:
 
         sub._received += 1
         if sub._max_msgs > 0 and sub._received >= sub._max_msgs:
-            # Enough messages so can throwaway subscription now.
+            # Enough messages so can throwaway subscription now, the
+            # pending messages will still be in the subscription
+            # internal queue and the task will finish once the last
+            # message is processed.
             self._subs.pop(sid, None)
-            sub._stop_processing()
 
         hdr: Optional[Dict[str, str]] = None
         if headers:
@@ -1820,9 +1870,9 @@ class Client:
             if not self.is_connected or self.is_connecting:
                 break
 
-            try:
-                await self._flush_queue.get()
+            future: asyncio.Future = await self._flush_queue.get()
 
+            try:
                 if self._pending_data_size > 0:
                     self._io_writer.writelines(self._pending[:])
                     self._pending = []
@@ -1835,6 +1885,8 @@ class Client:
             except (asyncio.CancelledError, RuntimeError, AttributeError):
                 # RuntimeError in case the event loop is closed
                 break
+            finally:
+                future.set_result(None)
 
     async def _ping_interval(self) -> None:
         while True:
@@ -1866,8 +1918,9 @@ class Client:
                 if should_bail or self._io_reader is None:
                     break
                 if self.is_connected and self._io_reader.at_eof():
-                    await self._error_cb(errors.StaleConnectionError())
-                    await self._process_op_err(errors.StaleConnectionError())
+                    err = errors.UnexpectedEOF()
+                    await self._error_cb(err)
+                    await self._process_op_err(err)
                     break
 
                 b = await self._io_reader.read(DEFAULT_BUFFER_SIZE)
